# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Pierre Verkest <pverkest@anybox.fr>
#    Copyright (C) 2019 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.testing import sgdb_in
import pytest
from anyblok import Declarations
from anyblok.mapper import ModelAttributeException
from anyblok.column import Integer, String, DateTime
from anyblok.relationship import Many2Many, Many2One
from anyblok.field import FieldException
from datetime import datetime
from .conftest import init_registry, reset_db

register = Declarations.register
Model = Declarations.Model
Mixin = Declarations.Mixin


def _complete_many2many(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        addresses = Many2Many(model=Model.Address,
                              join_table="join_addresses_by_persons",
                              remote_columns="id", local_columns="name",
                              m2m_remote_columns='a_id',
                              m2m_local_columns='p_name',
                              many2many="persons")


@pytest.fixture(scope="class")
def registry_many2many(request, bloks_loaded):
    reset_db()
    registry = init_registry(_complete_many2many)
    request.addfinalizer(registry.close)
    return registry


class TestMany2ManyComplete:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_many2many):
        transaction = registry_many2many.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_complete_many2many(self, registry_many2many):
        registry = registry_many2many

        address_exist = hasattr(registry.Person, 'addresses')
        assert address_exist

        m2m_tables_exist = hasattr(registry, 'many2many_tables')
        assert m2m_tables_exist

        jt = registry.declarativebase.metadata.tables
        join_table_exist = 'join_addresses_by_persons' in jt
        assert join_table_exist

        assert len(jt['join_addresses_by_persons'].primary_key.columns) == 2
        assert 'a_id' in jt[
            'join_addresses_by_persons'].primary_key.columns
        assert 'p_name' in jt[
            'join_addresses_by_persons'].primary_key.columns

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")

        person.addresses.append(address)

        assert address.persons == [person]

    def test_many2many_autodoc(self, registry_many2many):
        registry = registry_many2many
        registry.loaded_namespaces_first_step['Model.Person'][
            'addresses'].autodoc_get_properties()

    def test_complete_many2many_expire_field(self, registry_many2many):
        registry = registry_many2many
        assert ('x2m', 'addresses', 'persons') in registry.expire_attributes[
            'Model.Person']['addresses']
        assert ('x2m', 'persons', 'addresses') in registry.expire_attributes[
            'Model.Address']['persons']


def _minimum_many2many(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        addresses = Many2Many(model=Model.Address)


def _minimum_many2many_by_mixin(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Mixin)
    class MixinM2M:
        addresses = Many2Many(model=Model.Address)

    @register(Model)
    class Person(Mixin.MixinM2M):

        name = String(primary_key=True)


def _many2many_with_str_model(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        addresses = Many2Many(model='Model.Address')


def auto_detect_two_primary_keys(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        id2 = Integer(primary_key=True)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        addresses = Many2Many(model=Model.Address)


def unexisting_remote_columns(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        addresses = Many2Many(model=Model.Address, remote_columns="id2")


def reuse_many2many_table(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        addresses = Many2Many(model=Model.Address)

    @register(Model)  # noqa
    class Address:

        id = Integer(primary_key=True)
        persons = Many2Many(
            model=Model.Person,
            join_table='join_person_and_address_for_addresses')


class TestMany2Many:

    @pytest.fixture(autouse=True)
    def close_registry(self, request, bloks_loaded):

        def close():
            if hasattr(self, 'registry'):
                self.registry.close()

        request.addfinalizer(close)

    def init_registry(self, *args, **kwargs):
        reset_db()
        self.registry = init_registry(*args, **kwargs)
        return self.registry

    def test_minimum_many2many(self):
        registry = self.init_registry(_minimum_many2many)

        address_exist = hasattr(registry.Person, 'addresses')
        assert address_exist

        m2m_tables_exist = hasattr(registry, 'many2many_tables')
        assert m2m_tables_exist

        jt = registry.declarativebase.metadata.tables
        join_table_exist = 'join_person_and_address_for_addresses' in jt
        assert join_table_exist

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")

        person.addresses.append(address)

        assert person.addresses == [address]

    def test_many2many_with_str_model(self):
        registry = self.init_registry(_many2many_with_str_model)

        address_exist = hasattr(registry.Person, 'addresses')
        assert address_exist

        m2m_tables_exist = hasattr(registry, 'many2many_tables')
        assert m2m_tables_exist

        jt = registry.declarativebase.metadata.tables
        join_table_exist = 'join_person_and_address_for_addresses' in jt
        assert join_table_exist

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")

        person.addresses.append(address)

        assert person.addresses == [address]

    def test_unexisting_remote_columns(self):
        with pytest.raises(ModelAttributeException):
            self.init_registry(unexisting_remote_columns)

    def test_reuse_many2many_table(self):
        registry = self.init_registry(reuse_many2many_table)

        address = registry.Address.insert()
        person = registry.Person.insert(name="Jean-sébastien SUZANNE")

        person.addresses.append(address)

        assert address.persons == [person]

    def test_declared_in_mixin(self):
        registry = self.init_registry(_minimum_many2many_by_mixin)

        address_exist = hasattr(registry.Person, 'addresses')
        assert address_exist

        m2m_tables_exist = hasattr(registry, 'many2many_tables')
        assert m2m_tables_exist

        jt = registry.declarativebase.metadata.tables
        join_table_exist = 'join_person_and_address_for_addresses' in jt
        assert join_table_exist

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")

        person.addresses.append(address)

        assert person.addresses == [address]

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']), reason='ISSUE #90')
    def test_declared_in_mixin_inherit_by_two_models(self):
        def add_in_registry():
            _minimum_many2many_by_mixin()

            @register(Model)
            class Person2(Mixin.MixinM2M):

                name = String(primary_key=True)

        registry = self.init_registry(add_in_registry)

        address_exist = hasattr(registry.Person, 'addresses')
        assert address_exist
        address_exist = hasattr(registry.Person2, 'addresses')
        assert address_exist

        m2m_tables_exist = hasattr(registry, 'many2many_tables')
        assert m2m_tables_exist

        jt = registry.declarativebase.metadata.tables
        join_table_exist = 'join_person_and_address_for_addresses' in jt
        assert join_table_exist
        join_table_exist = 'join_person2_and_address_for_addresses' in jt
        assert join_table_exist

    def test_comlet_with_multi_primary_keys_remote_and_local(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)

            @register(Model)
            class Test2:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)
                test = Many2Many(model=Model.Test,
                                 join_table="join_test_and_test2",
                                 remote_columns=['id', 'id2'],
                                 m2m_remote_columns=['t2_id', 't2_id2'],
                                 local_columns=['id', 'id2'],
                                 m2m_local_columns=['t1_id', 't1_id2'],
                                 many2many="test2")

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert(id2="test1")
        t2 = registry.Test2.insert(id2="test2")
        t2.test.append(t1)
        assert t1.test2[0] is t2

        jt = registry.declarativebase.metadata.tables
        assert len(jt['join_test_and_test2'].primary_key.columns) == 4
        assert 't1_id' in jt[
            'join_test_and_test2'].primary_key.columns
        assert 't1_id2' in jt[
            'join_test_and_test2'].primary_key.columns
        assert 't2_id' in jt[
            'join_test_and_test2'].primary_key.columns
        assert 't2_id2' in jt[
            'join_test_and_test2'].primary_key.columns

    def test_comlet_with_multi_primary_keys_remote(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)

            @register(Model)
            class Test2:
                id = Integer(primary_key=True)
                test = Many2Many(model=Model.Test,
                                 join_table="join_test_and_test2",
                                 remote_columns=['id', 'id2'],
                                 m2m_remote_columns=['t2_id', 't2_id2'],
                                 local_columns='id',
                                 m2m_local_columns='t1_id',
                                 many2many="test2")

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert(id2="test1")
        t2 = registry.Test2.insert()
        t2.test.append(t1)
        assert t1.test2[0] is t2

    def test_comlet_with_multi_primary_keys_local(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)

            @register(Model)
            class Test2:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)
                test = Many2Many(model=Model.Test,
                                 join_table="join_test_and_test2",
                                 remote_columns='id',
                                 m2m_remote_columns='t2_id',
                                 local_columns=['id', 'id2'],
                                 m2m_local_columns=['t1_id', 't1_id2'],
                                 many2many="test2")

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test2.insert(id2='test2')
        t2.test.append(t1)
        assert t1.test2[0] is t2

    def test_minimum_with_multi_primary_keys_remote_and_local(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)

            @register(Model)
            class Test2:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)
                test = Many2Many(model=Model.Test, many2many="test2")

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert(id2="test1")
        t2 = registry.Test2.insert(id2="test2")
        t2.test.append(t1)
        assert t1.test2[0] is t2

    def test_minimum_with_multi_primary_keys_remote(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)

            @register(Model)
            class Test2:
                id = Integer(primary_key=True)
                test = Many2Many(model=Model.Test, many2many="test2")

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert(id2="test1")
        t2 = registry.Test2.insert()
        t2.test.append(t1)
        assert t1.test2[0] is t2

    def test_minimum_with_multi_primary_keys_local(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)

            @register(Model)
            class Test2:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)
                test = Many2Many(model=Model.Test, many2many="test2")

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test2.insert(id2="test2")
        t2.test.append(t1)
        assert t1.test2[0] is t2

    def test_many2many_on_self(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test',
                    m2m_remote_columns='id2',
                    many2many='parents'
                )

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert()
        t1.childs.append(t2)
        assert t1 in t2.parents

    def test_many2many_on_self_auto_column(self):

        def add_in_registry():

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test2',
                    many2many='parents'
                )

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test2.insert()
        t2 = registry.Test2.insert()
        t1.childs.append(t2)
        assert t1 in t2.parents

    def test_many2many_on_multi_fk_miss_on_m2m_column(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String(primary_key=True)

            @register(Model)
            class Person:

                name = String(primary_key=True)
                addresses = Many2Many(
                    m2m_remote_columns=['test_id'],
                    model=Model.Address)

        with pytest.raises(FieldException):
            self.init_registry(add_in_registry)

    def test_rich_many2many_complete_config(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class PersonAddress:
                id = Integer(primary_key=True)
                a_id = Integer(
                    foreign_key=Model.Address.use('id'), nullable=False)
                p_name = String(
                    foreign_key='Model.Person=>name', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Person:

                name = String(primary_key=True)
                addresses = Many2Many(model=Model.Address,
                                      join_table="personaddress",
                                      remote_columns="id", local_columns="name",
                                      m2m_remote_columns='a_id',
                                      m2m_local_columns='p_name',
                                      many2many="persons")

        registry = self.init_registry(add_in_registry)
        person = registry.Person.insert(name='jssuzanne')
        address = registry.Address.insert(
            street='somewhere', zip="75001", city="Paris")
        person.addresses.append(address)
        personaddress = registry.PersonAddress.query().one()
        assert personaddress.a_id == address.id
        assert personaddress.p_name == person.name
        assert personaddress.id
        assert personaddress.create_at
        assert personaddress.foo == 'bar'

    def test_rich_many2many_minimum_config(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class PersonAddress:
                id = Integer(primary_key=True)
                a_id = Integer(
                    foreign_key=Model.Address.use('id'), nullable=False)
                p_name = String(
                    foreign_key='Model.Person=>name', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Person:

                name = String(primary_key=True)
                addresses = Many2Many(model=Model.Address,
                                      join_table="personaddress",
                                      many2many="persons")

        registry = self.init_registry(add_in_registry)
        person = registry.Person.insert(name='jssuzanne')
        address = registry.Address.insert(
            street='somewhere', zip="75001", city="Paris")
        person.addresses.append(address)
        personaddress = registry.PersonAddress.query().one()
        assert personaddress.a_id == address.id
        assert personaddress.p_name == person.name
        assert personaddress.id
        assert personaddress.create_at
        assert personaddress.foo == 'bar'

    def test_rich_many2many_minimum_config_on_join_model(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class PersonAddress:
                id = Integer(primary_key=True)
                a_id = Integer(
                    foreign_key=Model.Address.use('id'), nullable=False)
                p_name = String(
                    foreign_key='Model.Person=>name', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Person:

                name = String(primary_key=True)
                addresses = Many2Many(model=Model.Address,
                                      join_model=Model.PersonAddress,
                                      many2many="persons")

        registry = self.init_registry(add_in_registry)
        person = registry.Person.insert(name='jssuzanne')
        address = registry.Address.insert(
            street='somewhere', zip="75001", city="Paris")
        person.addresses.append(address)
        personaddress = registry.PersonAddress.query().one()
        assert personaddress.a_id == address.id
        assert personaddress.p_name == person.name
        assert personaddress.id
        assert personaddress.create_at
        assert personaddress.foo == 'bar'

    def test_rich_many2many_minimum_config_with_many2one(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class PersonAddress:
                id = Integer(primary_key=True)
                person = Many2One(
                    model='Model.Person', nullable=False,
                    foreign_key_options={'ondelete': 'cascade'})
                address = Many2One(
                    model=Model.Address, nullable=False,
                    foreign_key_options={'ondelete': 'cascade'})
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Person:

                name = String(primary_key=True)
                addresses = Many2Many(model=Model.Address,
                                      join_model=Model.PersonAddress,
                                      many2many="persons")

        registry = self.init_registry(add_in_registry)
        person = registry.Person.insert(name='jssuzanne')
        address = registry.Address.insert(
            street='somewhere', zip="75001", city="Paris")
        person.addresses.append(address)
        personaddress = registry.PersonAddress.query().one()
        assert personaddress.address_id == address.id
        assert personaddress.person_name == person.name
        assert personaddress.id
        assert personaddress.create_at
        assert personaddress.foo == 'bar'

    def test_rich_many2many_minimum_config_with_pk_many2one(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class PersonAddress:
                person = Many2One(
                    model='Model.Person', nullable=False,
                    primary_key=True,
                    foreign_key_options={'ondelete': 'cascade'})
                address = Many2One(
                    model=Model.Address, nullable=False,
                    primary_key=True,
                    foreign_key_options={'ondelete': 'cascade'})
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Person:

                name = String(primary_key=True)
                addresses = Many2Many(model=Model.Address,
                                      join_model=Model.PersonAddress,
                                      many2many="persons")

        registry = self.init_registry(add_in_registry)
        person = registry.Person.insert(name='jssuzanne')
        address = registry.Address.insert(
            street='somewhere', zip="75001", city="Paris")
        person.addresses.append(address)
        personaddress = registry.PersonAddress.query().one()
        assert personaddress.address_id == address.id
        assert personaddress.person_name == person.name
        assert personaddress.create_at
        assert personaddress.foo == 'bar'

    def test_rich_many2many_minimum_config_on_join_model_and_join_table_1(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class PersonAddress:
                id = Integer(primary_key=True)
                a_id = Integer(
                    foreign_key=Model.Address.use('id'), nullable=False)
                p_name = String(
                    foreign_key='Model.Person=>name', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Person:

                name = String(primary_key=True)
                addresses = Many2Many(model=Model.Address,
                                      join_model=Model.PersonAddress,
                                      many2many="persons")

        registry = self.init_registry(add_in_registry)
        person = registry.Person.insert(name='jssuzanne')
        address = registry.Address.insert(
            street='somewhere', zip="75001", city="Paris")
        person.addresses.append(address)
        personaddress = registry.PersonAddress.query().one()
        assert personaddress.a_id == address.id
        assert personaddress.p_name == person.name
        assert personaddress.id
        assert personaddress.create_at
        assert personaddress.foo == 'bar'

    def test_rich_many2many_minimum_config_on_join_model_and_join_table_2(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class PersonAddress:
                id = Integer(primary_key=True)
                a_id = Integer(
                    foreign_key=Model.Address.use('id'), nullable=False)
                p_name = String(
                    foreign_key='Model.Person=>name', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Person:

                name = String(primary_key=True)
                addresses = Many2Many(model=Model.Address,
                                      join_table="other_join_table",
                                      join_model=Model.PersonAddress,
                                      many2many="persons")

        with pytest.raises(FieldException):
            self.init_registry(add_in_registry)

    def test_rich_many2many_complete_config_on_self(self):

        def add_in_registry():

            @register(Model)
            class TestLink:
                id = Integer(primary_key=True)
                t_left = Integer(foreign_key='Model.Test=>id', nullable=False)
                t_right = Integer(foreign_key='Model.Test=>id', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test',
                    many2many='parents',
                    join_table="testlink",
                    remote_columns="id", local_columns="id",
                    m2m_local_columns='t_left',
                    m2m_remote_columns='t_right',
                )

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert()
        t1.parents.append(t2)
        link = registry.TestLink.query().one()
        assert link.t_left == t2.id
        assert link.t_right == t1.id
        assert link.id
        assert link.create_at
        assert link.foo == 'bar'

    def test_rich_many2many_minimum_config_on_self(self):

        def add_in_registry():

            @register(Model)
            class TestLink:
                id = Integer(primary_key=True)
                t_left = Integer(foreign_key='Model.Test=>id', nullable=False)
                t_right = Integer(foreign_key='Model.Test=>id', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test',
                    many2many='parents',
                    join_table="testlink",
                    m2m_local_columns='t_left',
                    m2m_remote_columns='t_right',
                )

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert()
        t1.parents.append(t2)
        link = registry.TestLink.query().one()
        assert link.t_left == t2.id
        assert link.t_right == t1.id
        assert link.id
        assert link.create_at
        assert link.foo == 'bar'

    def test_rich_many2many_minimum_config_on_self_without_columns_1(self):

        def add_in_registry():

            @register(Model)
            class TestLink:
                id = Integer(primary_key=True)
                t_left = Integer(foreign_key='Model.Test=>id', nullable=False)
                t_right = Integer(foreign_key='Model.Test=>id', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test',
                    many2many='parents',
                    join_table="testlink",
                    m2m_local_columns='t_left',
                )

        with pytest.raises(FieldException):
            self.init_registry(add_in_registry)

    def test_rich_many2many_minimum_config_on_self_without_columns_2(self):

        def add_in_registry():

            @register(Model)
            class TestLink:
                id = Integer(primary_key=True)
                t_left = Integer(foreign_key='Model.Test=>id', nullable=False)
                t_right = Integer(foreign_key='Model.Test=>id', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test',
                    many2many='parents',
                    join_table="testlink",
                    m2m_remote_columns='t_right',
                )

        with pytest.raises(FieldException):
            self.init_registry(add_in_registry)

    def test_rich_many2many_minimum_config_on_self_with_join_model(self):

        def add_in_registry():

            @register(Model)
            class TestLink:
                id = Integer(primary_key=True)
                t_left = Integer(foreign_key='Model.Test=>id', nullable=False)
                t_right = Integer(foreign_key='Model.Test=>id', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test',
                    many2many='parents',
                    join_model=Model.TestLink,
                    m2m_local_columns='t_left',
                    m2m_remote_columns='t_right',
                )

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert()
        t1.parents.append(t2)
        link = registry.TestLink.query().one()
        assert link.t_left == t2.id
        assert link.t_right == t1.id
        assert link.id
        assert link.create_at
        assert link.foo == 'bar'

    def test_rich_many2many_minimum_config_on_self_with_many2one(self):

        def add_in_registry():

            @register(Model)
            class TestLink:
                id = Integer(primary_key=True)
                left = Many2One(
                    model='Model.Test', nullable=False,
                    foreign_key_options={'ondelete': 'cascade'})
                right = Many2One(
                    model='Model.Test', nullable=False,
                    foreign_key_options={'ondelete': 'cascade'})
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test',
                    many2many='parents',
                    join_model=Model.TestLink,
                    m2m_local_columns='left_id',
                    m2m_remote_columns='right_id',
                )

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert()
        t1.parents.append(t2)
        link = registry.TestLink.query().one()
        assert link.left_id == t2.id
        assert link.right_id == t1.id
        assert link.id
        assert link.create_at
        assert link.foo == 'bar'

    def test_rich_many2many_minimum_config_on_self_with_pk_many2one_2(self):

        def add_in_registry():

            @register(Model)
            class TestLink:
                left = Many2One(
                    model='Model.Test', nullable=False,
                    primary_key=True,
                    foreign_key_options={'ondelete': 'cascade'})
                right = Many2One(
                    model='Model.Test', nullable=False,
                    primary_key=True,
                    foreign_key_options={'ondelete': 'cascade'})
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test',
                    many2many='parents',
                    join_model=Model.TestLink,
                    m2m_local_columns='left',
                    m2m_remote_columns='right',
                )

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert()
        t1.parents.append(t2)
        link = registry.TestLink.query().one()
        assert link.left_id == t2.id
        assert link.right_id == t1.id
        assert link.create_at
        assert link.foo == 'bar'

    def test_with_twice_the_same_many2many(self):
        def add_in_registry(**kwargs):

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class Person:

                name = String(primary_key=True)
                invoiced_addresses = Many2Many(model=Model.Address)
                delivery_addresses = Many2Many(model=Model.Address)

        registry = self.init_registry(add_in_registry)
        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')
        person = registry.Person.insert(name="Jean-sébastien SUZANNE")

        person.invoiced_addresses.append(address)

        assert person.invoiced_addresses == [address]
        assert person.delivery_addresses == []



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from packaging.version import Version, LegacyVersion
import packaging


class VersionComparator:

    def parse_other(self, other):
        if other is None:
            other = LegacyVersion('')
        elif not isinstance(other, (Version, LegacyVersion)):
            other = parse_version(other)

        return other

    def __lt__(self, other):
        other = self.parse_other(other)
        return super(VersionComparator, self).__lt__(other)

    def __le__(self, other):
        other = self.parse_other(other)
        return super(VersionComparator, self).__le__(other)

    def __eq__(self, other):
        other = self.parse_other(other)
        return super(VersionComparator, self).__eq__(other)

    def __ne__(self, other):
        other = self.parse_other(other)
        return super(VersionComparator, self).__ne__(other)

    def __ge__(self, other):
        other = self.parse_other(other)
        return super(VersionComparator, self).__ge__(other)

    def __gt__(self, other):
        other = self.parse_other(other)
        return super(VersionComparator, self).__gt__(other)


class AnyBlokVersion(VersionComparator, Version):
    pass


class AnyBlokLegacyVersion(VersionComparator, LegacyVersion):
    pass


def parse_version(v):
    try:
        return AnyBlokVersion(v)
    except packaging.version.InvalidVersion:
        return AnyBlokLegacyVersion(v)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2016 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Pierre Verkest <pverkest@anybox.fr>
#    Copyright (C) 2019 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from anyblok.testing import sgdb_in
from anyblok.column import Integer as Int, String as Str
from anyblok.migration import MigrationException
from anyblok.relationship import Many2Many
from contextlib import contextmanager
from sqlalchemy import Column, Integer, TEXT, CheckConstraint
from anyblok import Declarations
from sqlalchemy.exc import InternalError, IntegrityError, OperationalError
from anyblok.config import Configuration
from .conftest import init_registry, drop_database, create_database


@pytest.fixture(scope="module")
def clean_db(request):
    def clean():
        url = Configuration.get('get_url')()
        drop_database(url)
        db_template_name = Configuration.get('db_template_name', None)
        create_database(url, template=db_template_name)

    request.addfinalizer(clean)


@contextmanager
def cnx(registry):
    cnx = registry.migration.conn
    try:
        yield cnx
    except Exception:
        cnx.execute("rollback")
        raise


def add_in_registry():
    register = Declarations.register
    Model = Declarations.Model

    @register(Model)
    class Test:
        integer = Int(primary_key=True)
        other = Str()

    @register(Model)
    class TestUnique:
        integer = Int(primary_key=True)
        other = Str(unique=True)

    @register(Model)
    class TestIndex:
        integer = Int(primary_key=True)
        other = Str(index=True)

    if not sgdb_in(['MySQL', 'MariaDB']):
        @register(Model)
        class TestCheck:
            integer = Int(primary_key=True)

            @classmethod
            def define_table_args(cls):
                table_args = super(TestCheck, cls).define_table_args()
                return table_args + (
                    CheckConstraint('integer > 0', name='test'),)

        @register(Model)
        class TestCheckLongConstraintName:
            integer = Int(primary_key=True)

            @classmethod
            def define_table_args(cls):
                table_args = super(TestCheckLongConstraintName,
                                   cls).define_table_args()
                return table_args + (
                    CheckConstraint('integer > 0', name=(
                        'long_long_long_long_long_long_long_long_long_long_'
                        'long_long_long_long_long_long_long_long_test')),)

    @register(Model)
    class TestFKTarget:
        integer = Int(primary_key=True)

    @register(Model)
    class TestFK:
        integer = Int(primary_key=True)
        other = Int(
            foreign_key=Model.TestFKTarget.use('integer'))

    @register(Model)
    class TestFK2:
        integer = Int(primary_key=True)
        other = Int(
            foreign_key=Model.TestFKTarget.use('integer').options(
                ondelete='cascade'))

    @register(Model)
    class TestM2M1:
        idmodel1 = Int(primary_key=True)

    @register(Model)
    class TestM2M2:
        idmodel2 = Int(primary_key=True)
        rel_m2m = Many2Many(label="Rel", model=Model.TestM2M1,
                            join_table='reltable',
                            remote_columns='idmodel1',
                            m2m_remote_columns='idmodel1',
                            local_columns='idmodel2',
                            m2m_local_columns='idmodel2',
                            many2many='rel_m2m_inv')


@pytest.fixture()
def registry(request, clean_db, bloks_loaded):
    registry = init_registry(add_in_registry)

    def rollback():
        for table in ('reltable', 'test', 'test2', 'othername', 'testfk',
                      'testunique', 'reltab', 'testm2m1',
                      'testm2m2', 'testfk2', 'testfktarget',  'testcheck',
                      'testchecklongconstraintname', 'testindex'):
            try:
                registry.migration.table(table).drop()
            except MigrationException:
                pass

        registry.close()

    request.addfinalizer(rollback)
    return registry


class TestMigration:

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Can't create empty table")
    def test_add_table(self, registry):
        registry.migration.table().add('test2')
        registry.migration.table('test2')

    def fill_test_table(self, registry):
        Test = registry.Test
        vals = [{'other': 'test %d' % x} for x in range(10)]
        Test.multi_insert(*vals)

    def test_add_column(self, registry):
        t = registry.migration.table('test')
        t.column().add(Column('new_column', Integer))
        t.column('new_column')

    def test_add_column_in_filled_table(self, registry):
        self.fill_test_table(registry)
        t = registry.migration.table('test')
        t.column().add(Column('new_column', Integer))
        t.column('new_column')

    def test_add_not_null_column(self, registry):
        t = registry.migration.table('test')
        t.column().add(Column('new_column', Integer, nullable=False))
        t.column('new_column')

    def test_add_not_null_column_in_filled_table(self, registry):
        self.fill_test_table(registry)
        t = registry.migration.table('test')
        t.column().add(Column('new_column', Integer, nullable=False))
        t.column('new_column')

    def test_add_column_with_default_value(self, registry):
        t = registry.migration.table('test')
        t.column().add(Column('new_column', Integer, default=100))
        t.column('new_column')

    def test_add_column_with_default_str_value(self, registry):
        t = registry.migration.table('test')
        t.column().add(Column('new_column', Integer, default='100'))
        t.column('new_column')

    def test_add_column_in_filled_table_with_default_value(self, registry):
        self.fill_test_table(registry)
        t = registry.migration.table('test')
        t.column().add(Column('new_column', Integer, default=100))
        t.column('new_column')
        res = [x for x in registry.execute(
            "select count(*) from test where new_column is null")][0][0]
        assert res == 0

    def test_add_not_null_column_in_filled_table_with_default_value(
        self, registry
    ):
        self.fill_test_table(registry)
        t = registry.migration.table('test')
        t.column().add(Column('new_column', Integer, nullable=False,
                       server_default="1"))
        t.column('new_column')

    def test_add_unique_constraint_on_good_table(self, registry):
        self.fill_test_table(registry)
        t = registry.migration.table('test')
        t.unique('unique_constraint').add(t.column('other'))
        registry.Test.insert(other='One entry')
        with pytest.raises(IntegrityError):
            registry.Test.insert(other='One entry')

    def test_add_unique_constraint_on_not_unique_column(self, registry):
        Test = registry.Test
        vals = [{'other': 'test'} for x in range(10)]
        Test.multi_insert(*vals)
        t = registry.migration.table('test')
        t.unique(None).add(t.column('other'))
        registry.Test.insert(other='One entry')
        registry.Test.insert(other='One entry')

    def test_drop_table(self, registry):
        registry.migration.table('test').drop()
        with pytest.raises(MigrationException):
            registry.migration.table('Test')

    def test_drop_column(self, registry):
        t = registry.migration.table('test')
        t.column('other').drop()
        with pytest.raises(MigrationException):
            t.column('other')

    def test_alter_table_name(self, registry):
        t = registry.migration.table('test')
        t.alter(name='othername')
        registry.migration.table('othername')
        with pytest.raises(MigrationException):
            registry.migration.table('test')

    def test_alter_column_name(self, registry):
        t = registry.migration.table('test')
        t.column('other').alter(name='othername')
        t.column('othername')
        with pytest.raises(MigrationException):
            t.column('other')

    def test_alter_column_nullable(self, registry):
        t = registry.migration.table('test')
        c = t.column('other').alter(nullable=False)
        assert not c.nullable

    def test_alter_column_nullable_in_filled_table(self, registry):
        t = registry.migration.table('test')
        t.column().add(Column('new_column', Integer))
        self.fill_test_table(registry)
        c = t.column('new_column').alter(nullable=False)
        # the column doesn't change of nullable to not lock the migration
        assert c.nullable

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for postgres")
    def test_alter_column_default(self, registry):
        t = registry.migration.table('test')
        c = t.column('other').alter(server_default='test')
        assert c.server_default == "'test'::character varying"

    def test_index(self, registry):
        t = registry.migration.table('test')
        t.index().add(t.column('integer'))
        t.index(t.column('integer')).drop()

    def test_alter_column_type(self, registry):
        t = registry.migration.table('test')
        c = t.column('other').alter(type_=TEXT)
        assert isinstance(c.type, TEXT)

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Can't drop primary key issue #92")
    def test_alter_column_primary_key(self, registry):
        t = registry.migration.table('test')
        t.primarykey().drop().add(t.column('other'))

    def test_alter_column_foreign_key(self, registry):
        t = registry.migration.table('test')
        t.foreign_key('my_fk').add(
            ['other'], [registry.System.Blok.name])
        t.foreign_key('my_fk').drop()

    def test_constraint_unique(self, registry):
        t = registry.migration.table('test')
        t.unique(name='test_unique_constraint').add(t.column('other'))
        t.unique(name='test_unique_constraint').drop()

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Can't drop check constraint issue #93")
    def test_constraint_check(self, registry):
        t = registry.migration.table('test')
        Test = registry.Test
        t.check('test').add(Test.other != 'test')
        # particuliar case of check constraint
        t.check('anyblok_ck_test__test').drop()

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_under_noautocommit_flag(self, registry):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) NOT NULL
                );""")
        registry.migration.detect_changed()
        registry.migration.withoutautomigration = True
        with pytest.raises(MigrationException):
            registry.migration.detect_changed()

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_column_added(self, registry):
        # Remove a column on the table force the detection to found new column
        # which is existing in metadata but not in table
        with cnx(registry) as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(integer INT PRIMARY KEY NOT NULL);""")
        report = registry.migration.detect_changed()
        assert report.log_has("Add test.other")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not report.log_has("Add test.other")

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_table_removed(self, registry):
        with cnx(registry) as conn:
            conn.execute(
                """CREATE TABLE test2(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64),
                    other2 CHAR(64)
                );""")

        report = registry.migration.detect_changed()
        assert report.log_has("Drop Table test2")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert report.log_has("Drop Table test2")

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_table_removed_with_reinit_column(self, registry):
        with cnx(registry) as conn:
            conn.execute(
                """CREATE TABLE test2(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64),
                    other2 CHAR(64)
                );""")

        registry.migration.reinit_tables = True
        report = registry.migration.detect_changed()
        assert report.log_has("Drop Table test2")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has("Drop Table test2"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_table_removed_with_reinit_all(self, registry):
        with cnx(registry) as conn:
            conn.execute(
                """CREATE TABLE test2(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64),
                    other2 CHAR(64)
                );""")

        registry.migration.reinit_all = True
        report = registry.migration.detect_changed()
        assert report.log_has("Drop Table test2")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has("Drop Table test2"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_column_removed(self, registry):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64),
                    other2 CHAR(64)
                );""")
        report = registry.migration.detect_changed()
        assert report.log_has("Drop Column test.other2")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert report.log_has("Drop Column test.other2")

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_column_removed_with_reinit_column(self, registry):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64),
                    other2 CHAR(64)
                );""")
        registry.migration.reinit_columns = True
        report = registry.migration.detect_changed()
        assert report.log_has("Drop Column test.other2")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has("Drop Column test.other2"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_column_removed_with_reinit_all(self, registry):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64),
                    other2 CHAR(64)
                );""")
        registry.migration.reinit_all = True
        report = registry.migration.detect_changed()
        assert report.log_has("Drop Column test.other2")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has("Drop Column test.other2"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_not_nullable_column_removed(self, registry):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64),
                    other2 CHAR(64) NOT NULL
                );""")
        report = registry.migration.detect_changed()
        assert not(report.log_has("Drop Column test.other2"))
        assert report.log_has("Drop Column test.other2 (not null)")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert report.log_has("Drop Column test.other2")
        assert not(report.log_has("Drop Column test.other2 (not null)"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_nullable(self, registry):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) NOT NULL
                );""")
        report = registry.migration.detect_changed()
        assert report.log_has("Alter test.other")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has("Alter test.other"))

    def test_detect_m2m_primary_key(self, registry):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE reltable")
            conn.execute(
                """CREATE TABLE reltable (
                    idmodel1 INT,
                    idmodel2 INT,
                    FOREIGN KEY (idmodel1) REFERENCES testm2m1 (idmodel1),
                    FOREIGN KEY (idmodel2) REFERENCES testm2m2 (idmodel2)
                );""")

        with pytest.raises(MigrationException):
            registry.migration.detect_changed()

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_server_default(self, registry):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) DEFAULT 9.99
                );""")
        report = registry.migration.detect_changed()
        assert report.log_has("Alter test.other")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has("Alter test.other"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_add_index_constrainte(self, registry):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE testindex")
            conn.execute(
                """CREATE TABLE testindex(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64)
                );""")
        report = registry.migration.detect_changed()
        assert report.log_has("Add index constraint on testindex (other)")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has(
            "Add index constraint on testindex (other)"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_add_column_with_index_constrainte(self, registry):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE testindex")
            conn.execute(
                """CREATE TABLE testindex(
                    integer INT PRIMARY KEY NOT NULL
                );""")
        report = registry.migration.detect_changed()
        assert report.log_has("Add testindex.other")
        assert report.log_has("Add index constraint on testindex (other)")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has("Add testindex.other"))
        assert not(report.log_has(
            "Add index constraint on testindex (other)"))

    def test_detect_drop_index(self, registry):
        with cnx(registry) as conn:
            conn.execute("""CREATE INDEX other_idx ON test (other);""")
        report = registry.migration.detect_changed()
        assert report.log_has("Drop index other_idx on test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert report.log_has("Drop index other_idx on test")

    def test_detect_drop_anyblok_index(self, registry):
        with cnx(registry) as conn:
            conn.execute(
                """CREATE INDEX anyblok_ix_test__other ON test (other);""")
        report = registry.migration.detect_changed()
        assert report.log_has("Drop index anyblok_ix_test__other on test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(
            report.log_has("Drop index anyblok_ix_test__other on test"))

    def test_detect_drop_index_with_reinit_indexes(self, registry):
        with cnx(registry) as conn:
            conn.execute("""CREATE INDEX other_idx ON test (other);""")
        registry.migration.reinit_indexes = True
        report = registry.migration.detect_changed()
        assert report.log_has("Drop index other_idx on test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has("Drop index other_idx on test"))

    def test_detect_drop_index_with_reinit_all(self, registry):
        with cnx(registry) as conn:
            conn.execute("""CREATE INDEX other_idx ON test (other);""")
        registry.migration.reinit_all = True
        report = registry.migration.detect_changed()
        assert report.log_has("Drop index other_idx on test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has("Drop index other_idx on test"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_type(self, registry):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other INT
                );""")
        report = registry.migration.detect_changed()
        assert report.log_has("Alter test.other")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has("Alter test.other"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_primary_key(self, registry):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT NOT NULL,
                    other CHAR(64) PRIMARY KEY NOT NULL
                );""")

        with pytest.raises(MigrationException):
            registry.migration.detect_changed()

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_add_foreign_key(self, registry):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE testfk")
            conn.execute(
                """CREATE TABLE testfk(
                    integer INT PRIMARY KEY NOT NULL,
                    other INT
                );""")
        report = registry.migration.detect_changed()
        assert report.log_has(
            "Add Foreign keys on (testfk.other) => (testfktarget.integer)")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has(
            "Add Foreign keys on (testfk.other) => (testfktarget.integer)"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_foreign_key_options_changed(self, registry):
        with cnx(registry) as conn:
            conn.execute("drop table testfk2")
            conn.execute(
                """create table testfk2(
                    integer INT PRIMARY KEY NOT NULL,
                    other int
                        CONSTRAINT anyblok_fk_testfk2__other
                        references testfktarget(integer)
                );""")
        report = registry.migration.detect_changed()
        assert report.log_has(
            "Drop Foreign keys on testfk2.other => testfktarget.integer")
        assert report.log_has(
            "Add Foreign keys on (testfk2.other) => (testfktarget.integer)")

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_drop_foreign_key(self, registry):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) references system_blok(name)
                );""")
        report = registry.migration.detect_changed()
        assert report.log_has(
            "Drop Foreign keys on test.other => system_blok.name")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert report.log_has(
            "Drop Foreign keys on test.other => system_blok.name")

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_drop_anyblok_foreign_key(self, registry):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64)
                        CONSTRAINT anyblok_fk_test__other_on_system_blok__name
                        references system_blok(name)
                );""")
        report = registry.migration.detect_changed()
        assert report.log_has(
            "Drop Foreign keys on test.other => system_blok.name")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has(
            "Drop Foreign keys on test.other => system_blok.name"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_drop_foreign_key_with_reinit_constraint(self, registry):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) references system_blok(name)
                );""")
        registry.migration.reinit_constraints = True
        report = registry.migration.detect_changed()
        assert report.log_has(
            "Drop Foreign keys on test.other => system_blok.name")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has(
            "Drop Foreign keys on test.other => system_blok.name"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_drop_foreign_key_with_reinit_all(self, registry):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) references system_blok(name)
                );""")
        registry.migration.reinit_all = True
        report = registry.migration.detect_changed()
        assert report.log_has(
            "Drop Foreign keys on test.other => system_blok.name")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has(
            "Drop Foreign keys on test.other => system_blok.name"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_drop_column_with_foreign_key(self, registry):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64),
                    other2 CHAR(64)
                    CONSTRAINT anyblok_fk_test__other2_on_system_blok__name
                    references system_blok(name)
                );""")
        report = registry.migration.detect_changed()
        assert report.log_has(
            "Drop Foreign keys on test.other2 => system_blok.name")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has(
            "Drop Foreign keys on test.other2 => system_blok.name"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_add_unique_constrainte(self, registry):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE testunique")
            conn.execute(
                """CREATE TABLE testunique(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64)
                );""")
        report = registry.migration.detect_changed()
        assert report.log_has("Add unique constraint on testunique (other)")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has(
            "Add unique constraint on testunique (other)"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_add_column_with_unique_constrainte(self, registry):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE testunique")
            conn.execute(
                """CREATE TABLE testunique(
                    integer INT PRIMARY KEY NOT NULL
                );""")
        report = registry.migration.detect_changed()
        assert report.log_has("Add unique constraint on testunique (other)")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has(
            "Add unique constraint on testunique (other)"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_drop_unique_constraint(self, registry):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) CONSTRAINT unique_other UNIQUE
                );""")
        report = registry.migration.detect_changed()
        assert report.log_has("Drop constraint unique_other on test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert report.log_has("Drop constraint unique_other on test")

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_drop_unique_anyblok_constraint(self, registry):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) CONSTRAINT anyblok_uq_test__other UNIQUE
                );""")
        report = registry.migration.detect_changed()
        assert report.log_has("Drop constraint anyblok_uq_test__other on test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(
            report.log_has("Drop constraint anyblok_uq_test__other on test"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_drop_unique_constraint_with_reinit_constraints(
        self, registry
    ):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) CONSTRAINT unique_other UNIQUE
                );""")

        registry.migration.reinit_constraints = True
        report = registry.migration.detect_changed()
        assert report.log_has("Drop constraint unique_other on test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(
            report.log_has("Drop constraint unique_other on test"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_drop_unique_constraint_with_reinit_all(self, registry):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) CONSTRAINT unique_other UNIQUE
                );""")

        registry.migration.reinit_all = True
        report = registry.migration.detect_changed()
        assert report.log_has("Drop constraint unique_other on test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(
            report.log_has("Drop constraint unique_other on test"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="No CheckConstraint works #90")
    def test_no_detect_drop_and_add_check_constraint_with_long_name(
        self, registry
    ):
        report = registry.migration.detect_changed()
        assert not(
            report.log_has(
                "Drop check constraint anyblok_ck_testchecklongconstraintname"
                "__long_long_long_long_lon on testchecklongconstraintname"
            )
        )
        assert not(
            report.log_has(
                "Add check constraint anyblok_ck_testchecklongconstraintname__"
                "long_long_long_long_long_long_long_long_long_long_long_long_"
                "long_long_long_long_long_long_test on "
                "testchecklongconstraintname"
            )
        )

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_add_check_constraint(self, registry):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE testcheck")
            conn.execute(
                """CREATE TABLE testcheck(
                    integer INT PRIMARY KEY NOT NULL
                );""")
        report = registry.migration.detect_changed()
        assert report.log_has(
            "Add check constraint anyblok_ck_testcheck__test on testcheck")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has(
            "Add check constraint anyblok_ck_testcheck__test on testcheck"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_drop_check_constraint(self, registry):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) CONSTRAINT ck_other CHECK (other != 'test')
                );""")
        report = registry.migration.detect_changed()
        assert report.log_has("Drop check constraint ck_other on test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert report.log_has("Drop check constraint ck_other on test")

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_drop_check_anyblok_constraint(self, registry):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) CONSTRAINT anyblok_ck__test__check
                        CHECK (other != 'test')
                );""")
        report = registry.migration.detect_changed()
        assert report.log_has(
            "Drop check constraint anyblok_ck__test__check on test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(report.log_has(
            "Drop check constraint anyblok_ck__test__check on test"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_drop_check_constraint_with_reinit_constraint(
        self, registry
    ):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) CONSTRAINT ck_other CHECK (other != 'test')
                );""")
        registry.migration.reinit_constraints = True
        report = registry.migration.detect_changed()
        assert report.log_has("Drop check constraint ck_other on test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(
            report.log_has("Drop check constraint ck_other on test"))

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB']),
                        reason="Test for Postgres only")
    def test_detect_drop_check_constraint_with_reinit_all(self, registry):
        with cnx(registry) as conn:
            conn.execute("DROP TABLE test")
            conn.execute(
                """CREATE TABLE test(
                    integer INT PRIMARY KEY NOT NULL,
                    other CHAR(64) CONSTRAINT ck_other CHECK (other != 'test')
                );""")
        registry.migration.reinit_all = True
        report = registry.migration.detect_changed()
        assert report.log_has("Drop check constraint ck_other on test")
        report.apply_change()
        report = registry.migration.detect_changed()
        assert not(
            report.log_has("Drop check constraint ck_other on test"))

    def test_savepoint(self, registry):
        Test = registry.Test
        self.fill_test_table(registry)
        registry.migration.savepoint('test')
        self.fill_test_table(registry)
        assert Test.query().count() == 20
        registry.migration.rollback_savepoint('test')
        assert Test.query().count() == 10
        registry.migration.release_savepoint('test')

    def test_savepoint_without_rollback(self, registry):
        registry.migration.savepoint('test')
        registry.migration.release_savepoint('test')
        with pytest.raises((InternalError, OperationalError)):
            registry.migration.rollback_savepoint('test')



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Denis VIVIÈS <dvivies@geoblink.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import logging
from copy import deepcopy
import pytest
from anyblok.testing import sgdb_in
from anyblok.blok import BlokManager
from anyblok.config import Configuration
from anyblok.environment import EnvironmentManager
from anyblok.registry import RegistryManager
from copy import copy
import sqlalchemy
from sqlalchemy_utils.functions import database_exists, create_database, orm
import os

logger = logging.getLogger(__name__)


def init_registry_with_bloks(bloks, function, **kwargs):
    if bloks is None:
        bloks = []
    if isinstance(bloks, tuple):
        bloks = list(bloks)
    if isinstance(bloks, str):
        bloks = [bloks]

    anyblok_test_name = 'anyblok-test'
    if anyblok_test_name not in bloks:
        bloks.append(anyblok_test_name)

    loaded_bloks = deepcopy(RegistryManager.loaded_bloks)
    if function is not None:
        EnvironmentManager.set('current_blok', anyblok_test_name)
        try:
            function(**kwargs)
        finally:
            EnvironmentManager.set('current_blok', None)
    try:
        registry = RegistryManager.get(
            Configuration.get('db_name'),
            unittest=True)

        # update required blok
        registry_bloks = registry.get_bloks_by_states('installed', 'toinstall')
        toinstall = [x for x in bloks if x not in registry_bloks]
        toupdate = [x for x in bloks if x in registry_bloks]
        registry.upgrade(install=toinstall, update=toupdate)
    finally:
        RegistryManager.loaded_bloks = loaded_bloks

    return registry


def init_registry(function, **kwargs):
    return init_registry_with_bloks([], function, **kwargs)


def drop_database(url):
    url = copy(sqlalchemy.engine.url.make_url(url))
    database = url.database
    if url.drivername.startswith('postgresql'):
        url.database = 'postgres'
    elif not url.drivername.startswith('sqlite'):
        url.database = None

    engine = sqlalchemy.create_engine(url)
    if engine.dialect.name == 'sqlite' and url.database != ':memory:':
        os.remove(url.database)
    else:
        text = 'DROP DATABASE {0}'.format(orm.quote(engine, database))
        cnx = engine.connect()
        cnx.execute("ROLLBACK")
        cnx.execute(text)
        cnx.execute("commit")
        cnx.close()


@pytest.fixture(scope="session")
def base_loaded(request, configuration_loaded):
    if sgdb_in(['MySQL', 'MariaDB']):
        return

    url = Configuration.get('get_url')()
    if not database_exists(url):
        db_template_name = Configuration.get('db_template_name', None)
        create_database(url, template=db_template_name)

    BlokManager.load()
    registry = init_registry_with_bloks([], None)
    registry.commit()
    registry.close()
    BlokManager.unload()


@pytest.fixture(scope="module")
def bloks_loaded(request, base_loaded):
    request.addfinalizer(BlokManager.unload)
    BlokManager.load()


@pytest.fixture(scope="module")
def testbloks_loaded(request, base_loaded):
    request.addfinalizer(BlokManager.unload)
    BlokManager.load(entry_points=('bloks', 'test_bloks'))


def reset_db():
    if sgdb_in(['MySQL', 'MariaDB']):
        url = Configuration.get('get_url')()
        if database_exists(url):
            drop_database(url)

        db_template_name = Configuration.get('db_template_name', None)
        create_database(url, template=db_template_name)


@pytest.fixture(scope="class")
def registry_blok(request, bloks_loaded):
    reset_db()
    registry = init_registry_with_bloks([], None)
    request.addfinalizer(registry.close)
    return registry


@pytest.fixture(scope="class")
def registry_testblok(request, testbloks_loaded):
    reset_db()
    registry = init_registry_with_bloks([], None)
    request.addfinalizer(registry.close)
    return registry


@pytest.fixture(scope="function")
def registry_testblok_func(request, testbloks_loaded):
    reset_db()
    registry = init_registry_with_bloks([], None)
    request.addfinalizer(registry.close)
    return registry



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest  # noqa
from anyblok.blok import BlokManager
from .conftest import init_registry


@pytest.fixture(scope="module")
def bloks_loaded(request):
    request.addfinalizer(BlokManager.unload)
    BlokManager.load()


class TestCoreQuery:

    @pytest.fixture(autouse=True)
    def close_registry(self, request, bloks_loaded):

        def close():
            if hasattr(self, 'registry'):
                self.registry.close()

        request.addfinalizer(close)

    def init_registry(self, *args, **kwargs):
        self.registry = init_registry(*args, **kwargs)
        return self.registry

    def test_inherit(self):

        def inherit():

            from anyblok import Declarations
            Core = Declarations.Core

            @Declarations.register(Core)
            class Query:

                def foo(self):
                    return True

        registry = self.init_registry(inherit)
        assert registry.System.Blok.query().foo() is True



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from anyblok.field import Field, FieldException, Function, JsonRelated
from anyblok.column import Integer, String, Json
from anyblok import Declarations
from sqlalchemy import func, types
from .conftest import init_registry


Model = Declarations.Model
register = Declarations.register


class OneField(Field):
    pass


class TestField:

    def test_forbid_instance(self):
        with pytest.raises(FieldException):
            Field()

    def test_without_label(self):
        field = OneField()
        field.get_sqlalchemy_mapping(None, None, 'a_field', None)
        assert field.label == 'A field'


def define_field_function():

    @register(Model)
    class Test:

        id = Integer(primary_key=True)
        first_name = String()
        last_name = String()
        name = Function(fget='fget', fset='fset', fdel='fdel',
                        fexpr='fexpr')

        def fget(self):
            return '{0} {1}'.format(self.first_name, self.last_name)

        def fset(self, value):
            self.first_name, self.last_name = value.split(' ', 1)

        def fdel(self):
            self.first_name = self.last_name = None

        @classmethod
        def fexpr(cls):
            return func.concat(cls.first_name, ' ', cls.last_name)


@pytest.fixture(scope="class")
def registry_field_function(request, bloks_loaded):
    registry = init_registry(define_field_function)
    request.addfinalizer(registry.close)
    return registry


class TestFieldFunction:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_field_function):
        transaction = registry_field_function.begin_nested()
        request.addfinalizer(transaction.rollback)

    def test_field_function_fget(self, registry_field_function):
        registry = registry_field_function
        t = registry.Test.insert(first_name='Jean-Sebastien',
                                 last_name='SUZANNE')
        assert t.name == 'Jean-Sebastien SUZANNE'
        t = registry.Test.query().first()
        assert t.name == 'Jean-Sebastien SUZANNE'

    def test_field_function_fdel(self, registry_field_function):
        registry = registry_field_function
        t = registry.Test.insert(first_name='jean-sebastien',
                                 last_name='suzanne')
        del t.name
        assert t.first_name is None
        assert t.last_name is None

    def test_field_function_fexpr(self, registry_field_function):
        registry = registry_field_function
        registry.Test.insert(first_name='Jean-Sebastien',
                             last_name='SUZANNE')
        t = registry.Test.query().filter(
            registry.Test.name == 'Jean-Sebastien SUZANNE').first()
        assert t.name == 'Jean-Sebastien SUZANNE'


def define_field_json_related():

    @register(Model)
    class Test:

        id = Integer(primary_key=True)
        properties = Json()
        name = JsonRelated(json_column='properties', keys=['name'])


@pytest.fixture(scope="class")
def registry_json_related(request, bloks_loaded):
    registry = init_registry(define_field_json_related)
    request.addfinalizer(registry.close)
    return registry


class TestJsonRelated:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_json_related):
        transaction = registry_json_related.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_field_json_related_hasattr(self, registry_json_related):
        registry = registry_json_related
        registry.Test.name

    def test_field_json_related_autodoc(self, registry_json_related):
        registry = registry_json_related
        registry.loaded_namespaces_first_step['Model.Test'][
            'name'].autodoc_get_properties()

    def test_field_json_related_get_1(self, registry_json_related):
        registry = registry_json_related
        t = registry.Test.insert()
        assert t.name is None

    def test_field_json_related_get_2(self, registry_json_related):
        registry = registry_json_related
        t = registry.Test.insert(properties={})
        assert t.name is None

    def test_field_json_related_get_3(self, registry_json_related):
        registry = registry_json_related
        t = registry.Test.insert(properties={'name': 'jssuzanne'})
        assert t.name == 'jssuzanne'

    def test_field_json_related_del_1(self, registry_json_related):
        registry = registry_json_related
        t = registry.Test.insert(properties={'name': 'jssuzanne'})
        del t.name
        assert t.name is None
        assert t.properties == {'name': None}

    def test_field_json_related_set_1(self, registry_json_related):
        registry = registry_json_related
        t = registry.Test.insert()
        t.name = 'jssuzanne'
        assert t.properties == {'name': 'jssuzanne'}

    def test_field_json_related_set_2(self, registry_json_related):
        registry = registry_json_related
        t = registry.Test.insert(properties={})
        t.name = 'jssuzanne'
        assert t.properties == {'name': 'jssuzanne'}

    def test_field_json_related_set_3(self, registry_json_related):
        registry = registry_json_related
        t = registry.Test.insert(properties={'name': 'other'})
        t.name = 'jssuzanne'
        assert t.properties == {'name': 'jssuzanne'}

    def test_field_json_related_exp_1(self, registry_json_related):
        registry = registry_json_related
        Test = registry.Test
        Test.insert()
        query = registry.Test.query().filter(
            Test.name.cast(types.String) == '"jssuzanne"')
        assert not (query.count())

    def test_field_json_related_exp_2(self, registry_json_related):
        registry = registry_json_related
        Test = registry.Test
        Test.insert(properties={})
        query = registry.Test.query().filter(
            Test.name.cast(types.String) == '"jssuzanne"')
        assert not (query.count())

    def test_field_json_related_exp_3(self, registry_json_related):
        registry = registry_json_related
        Test = registry.Test
        Test.insert(properties={'name': 'jssuzanne'})
        query = registry.Test.query().filter(
            Test.name.cast(types.String) == '"jssuzanne"')
        assert query.count()


def define_field_json_related2():

    @register(Model)
    class Test:

        id = Integer(primary_key=True)
        properties = Json()
        name = JsonRelated(json_column='properties', keys=['sub', 'name'])


@pytest.fixture(scope="class")
def registry_json_related2(request, bloks_loaded):
    registry = init_registry(define_field_json_related2)
    request.addfinalizer(registry.close)
    return registry


class TestJsonRelated2:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_json_related2):
        transaction = registry_json_related2.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_field_json_related_get_4(self, registry_json_related2):
        registry = registry_json_related2
        t = registry.Test.insert(
            properties={'sub': {'name': 'jssuzanne'}})
        assert t.name == 'jssuzanne'

    def test_field_json_related_del_2(self, registry_json_related2):
        registry = registry_json_related2
        t = registry.Test.insert()
        del t.name
        assert t.name is None
        assert t.properties == {'sub': {'name': None}}

    def test_field_json_related_set_4(self, registry_json_related2):
        registry = registry_json_related2
        t = registry.Test.insert(properties={'sub': {'name': 'other'}})
        t.name = 'jssuzanne'
        assert t.properties == {'sub': {'name': 'jssuzanne'}}

    def test_field_json_related_exp_4(self, registry_json_related2):
        registry = registry_json_related2
        Test = registry.Test
        Test.insert(properties={'sub': {'name': 'jssuzanne'}})
        query = registry.Test.query().filter(
            Test.name.cast(types.String) == '"jssuzanne"')
        assert query.count()


def field_without_name():

    @register(Model)
    class Test:

        id = Integer(primary_key=True)
        field = OneField()


class TestField2:

    @pytest.fixture(autouse=True)
    def close_registry(self, request, bloks_loaded):

        def close():
            if hasattr(self, 'registry'):
                self.registry.close()

        request.addfinalizer(close)

    def init_registry(self, *args, **kwargs):
        self.registry = init_registry(*args, **kwargs)
        return self.registry

    def test_field_without_name(self):
        self.init_registry(field_without_name)

    def test_field_function_fset(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                _name = String()
                name = Function(fget='fget', fset='fset', fdel='fdel',
                                fexpr='fexpr')

                def fget(self):
                    return self._name

                def fset(self, value):
                    self._name = value

                def fdel(self):
                    self._name = None

                @classmethod
                def fexpr(cls):
                    return cls._name

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert(name='Jean-Sebastien SUZANNE')
        assert t._name == 'Jean-Sebastien SUZANNE'
        t = registry.Test.query().first()
        t.name = 'Mister ANYBLOK'
        assert t._name == 'Mister ANYBLOK'
        t.update(name='Jean-Sebastien SUZANNE')
        assert t._name == 'Jean-Sebastien SUZANNE'

    def test_class_hasattr(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                val1 = Integer()
                val2 = Function(fget='fget', fset='fset')

                def fget(self):
                    return 2 * self.val1

        registry = self.init_registry(add_in_registry)
        registry.Test.val2

    def test_field_json_related_with_missing_json_column(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                properties = Json()
                name = JsonRelated(keys=['name'])

        with pytest.raises(FieldException):
            self.init_registry(add_in_registry)

    def test_field_json_related_with_missing_keys(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                properties = Json()
                name = JsonRelated(json_column='properties')

        with pytest.raises(FieldException):
            self.init_registry(add_in_registry)

    def test_field_json_related_with_adapter(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                properties = Json()
                name = JsonRelated(json_column='properties', keys=['name'],
                                   set_adapter=str, get_adapter=int)

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        assert t.name is None
        t.name = 1
        assert t.properties == {'name': '1'}
        t.properties['name'] = '2'
        assert t.name == 2

    def test_field_json_related_with_adapter_with_method(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                properties = Json()
                name = JsonRelated(json_column='properties', keys=['name'],
                                   set_adapter='fromint', get_adapter='toint')

                def fromint(self, value):
                    if value:
                        return str(value)

                    return value

                def toint(self, value):
                    if value:
                        return int(value)

                    return value

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        assert t.name is None
        t.name = 1
        assert t.properties == {'name': '1'}
        t.properties['name'] = '2'
        assert t.name == 2



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Denis VIVIÈS <dvivies@geoblink.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest  # noqa
from anyblok._graphviz import ModelSchema, SQLSchema
from anyblok.config import Configuration


class TestDocumentation:

    def test_autodoc(self, rollback_registry):
        registry = rollback_registry
        doc = registry.Documentation()
        doc.auto_doc()

    def test_to_rst(self, rollback_registry):
        registry = rollback_registry
        doc = registry.Documentation()
        doc.auto_doc()
        with open('test_doc_output', 'w') as fp:
            doc.toRST(fp)

    def test_to_uml(self, rollback_registry):
        registry = rollback_registry
        doc = registry.Documentation()
        doc.auto_doc()
        format_ = Configuration.get('schema_format')
        dot = ModelSchema('test', format=format_)
        doc.toUML(dot)
        dot.save()

    def test_to_sql(self, rollback_registry):
        registry = rollback_registry
        doc = registry.Documentation()
        doc.auto_doc()
        format_ = Configuration.get('schema_format')
        dot = SQLSchema('test', format=format_)
        doc.toSQL(dot)
        dot.save()
